Require Import Metatheory.
Require Import FJ_Definitions.
Require Import CFJ_Definitions.
Require Import CFJ_Facts.
Require Import Coq.Logic.ClassicalFacts.

(********************************************************************)
(* REACHABILITY PROOFS                                              *)
(********************************************************************)

Lemma class_superclass : (*$\label{class_superclass}$*)
  forall (C D:cname) (gs hs ts fs:a_flds) (ms:a_mths),
    binds C (AT C,D,gs,hs,ts,fs,ms) aCT ->
    (AT C -> AT D).
Proof.
  intros C D gs hs ts fs ms Hb.
  set aCT_well_typed as Hct.
  destruct Hct as [Hok Hfa].
  induction Hfa.
    discriminate Hb.

    destruct (x == C); subst.
      destruct v as [[[[[[AT_C D'] gs'] hs'] ts'] fs'] ms'].
      inversion Hb.
      rewrite eq_atom_true in H1.
      inversion H1; subst; clear H1.
      destruct H as [_ [Hr _]]; assumption.

      apply IHHfa.
        unfold binds in Hb; simpl in Hb.
        rewrite eq_atom_false in Hb; auto.

        inversion Hok; auto.
Qed.

Lemma field_class : (*$\label{field_class}$*)
  forall (C:cname) (f:fname), ATf C f -> AT C.
Proof.
  intros C f Hf.
  set aCT_well_typed as Hct.
  destruct Hct as [Hok Hfa].
  unfold ATf in Hf.
  induction Hfa.
    simpl in Hf.
    contradict Hf.

    destruct (x == C); subst.
      unfold a_ok_class' in H.
      destruct v as [[[[[[AT_C D] gs] hs] ts] fs] ms].
      unfold a_ok_class in H.
      destruct H as [[gs' [_ [Hokcc _]]] _].
      simpl in Hf.
      rewrite eq_atom_true in Hf.
      induction Hokcc; eauto.
        unfold get in Hf.
        contradict Hf; assumption.

        destruct (fi == f); subst.
        apply H.
        simpl in Hf.
        rewrite eq_atom_true in Hf; assumption.
          
        apply IHHokcc; eauto.
          simpl in Hf.
          rewrite eq_atom_false in Hf; eauto.
        
          inversion Hok; apply ok_cons; eauto.

        apply IHHokcc; eauto.
        inversion Hok; apply ok_cons; eauto.

      apply IHHfa.
        unfold get in Hf.
        rewrite eq_atom_false in Hf; eauto.

        inversion Hok; assumption.
Qed.

Lemma variable_present : (*$\label{variable_present}$*)
  forall (x:var) (a:ann) (t:typ) (E:a_env) (E':env),
    binds x (a,t) E ->
    a ->
    variant_env E E' ->
    binds x t E'.
Proof.
  intros x a t E E' Hnb Hr Hve.
  induction Hve.
    discriminate Hnb.

    unfold binds in *.
    simpl in Hnb.
    destruct (x == x0).
      inversion Hnb; subst.
      apply binds_first.

      apply binds_other; try eauto.
      apply IHHve; assumption.

    apply IHHve.
    unfold binds in *.
    simpl in Hnb.
    destruct (x == x0).
      inversion Hnb; subst.
      contradict Hr; assumption.

      assumption.
Qed.

Lemma class_present : (*$\label{class_present}$*)
  forall (C D:cname) (gs hs ts fs:a_flds) (ms:a_mths),
    AT C ->
    binds C (AT C,D,gs,hs,ts,fs,ms) aCT ->
    exists fs':flds, exists ms':mths,
      variant_env fs fs' /\
      variant_methods C ms ms' /\
      binds C (D,fs',ms') CT.
Proof.
  intros C D gs hs ts fs ms HrC Hb.
  set aCT_correlates_with_CT as Hct.
  induction Hct.
    discriminate Hb.

    destruct (C0 == C); subst.
      set (He := Hb).
      apply binds_elim_eq in He.
      symmetry in He.
      inversion He; clear He; subst.
      exists fs'; exists ms'.
      repeat split; try assumption.
      apply binds_first.

      unfold binds.
      simpl.
      rewrite eq_atom_false; eauto.
      apply IHHct.
      inversion Hb; rewrite eq_atom_false in H3; eauto.

    destruct (C0 == C); subst.
      contradict HrC; tauto.

      apply IHHct.
      unfold binds in Hb.
      simpl in Hb.
      rewrite eq_atom_false in Hb; eauto.
Qed.

Lemma class_present_reverse : (*$\label{class_present_reverse}$*)
  forall (C D:cname) (fs':flds) (ms':mths),
    binds C (D,fs',ms') CT ->
    exists gs:a_flds, exists hs:a_flds, exists ts:a_flds,
    exists fs:a_flds, exists ms:a_mths,
      variant_env fs fs' /\
      variant_methods C ms ms' /\
      binds C (AT C,D,gs,hs,ts,fs,ms) aCT.
Proof.
  intros C D fs' ms' Hb.
  set aCT_correlates_with_CT as Hct.
  assert (HrC:AT C).
    induction Hct; auto.
      discriminate Hb.

      destruct (C0 == C); subst. auto.
      inversion Hb.
      rewrite eq_atom_false in H3; auto.
      
    induction Hct.
      discriminate Hb.

      destruct (C0 == C); subst.
        set (He := Hb).
        apply binds_elim_eq in He.
        symmetry in He.
        inversion He; clear He; subst.
        exists gs; exists hs; exists ts; exists fs; exists ms.
        repeat split; auto.
        apply binds_first.

        unfold binds; simpl.
        rewrite eq_atom_false; eauto.
        apply IHHct.
        inversion Hb; rewrite eq_atom_false in H3; eauto.

      destruct (C0 == C); subst.
        contradict H; auto.

        unfold binds; simpl.
        rewrite eq_atom_false; auto.
Qed.

Lemma fields_present : (*$\label{fields_present}$*)
  forall (C:cname) (fs:a_flds),
    AT C ->
    a_fields C fs ->
    exists fs':flds, variant_env fs fs' /\ fields C fs'.
Proof.
  intros C fs0 HrC Hfs.
  set aCT_well_typed as Hct.
  destruct Hct as [Hok Hfa].
  induction Hfs.
    exists  nil; split; [apply ve_nil | apply fields_obj].

    assert (exists fs':flds, exists ms':mths, variant_env fs fs' /\
      variant_methods C ms ms' /\ binds C (D,fs',ms') CT)
      as [fs1 [ms1 [Hve1 [Hvm1 Hb1]]]]
      by (apply class_present with (gs:=gs) (hs:=hs) (ts:=ts); eauto).
    assert (HrCD:AT C -> AT D)
      by (apply class_superclass with (C:=C) (gs:=gs) (hs:=hs)
         (ts:=ts) (fs:=fs) (ms:=ms); eauto).
    assert (HrD:AT D) by auto.
    apply IHHfs in HrD.
    destruct HrD as [fs2 [Hve2 HD]].
    exists (fs2++fs1).
    split.
      revert Hve1; revert Hve2; clear. 
      induction 1; intro Hgs; repeat rewrite <- app_comm_cons; auto.

      apply fields_other with (D:=D) (ms:=ms1); eauto.
Qed.

Lemma field_present : (*$\label{field_present}$*)
  forall (C D:cname) (f:fname),
    ATf C f ->
    a_field C f (ATf C f) D ->
    field C f D.
Proof.
  intros C D f Hr Hf.
  inversion_clear Hf as [fs Hfs Hb]; subst.
  cut (AT C); intros;
    [idtac | apply field_class with (f:=f); eauto].
  assert (exists fs':flds, variant_env fs fs' /\ fields C fs') by
    (apply fields_present; eauto).
  destruct H0 as [fs' [H0 H1]].
  unfold field.
  exists fs'; try assumption.
  clear Hfs. clear H1.
  induction H0.
    discriminate Hb.

    destruct (x == f); subst.
      inversion Hb.
      rewrite eq_atom_true in H3.
      inversion H3; subst.
      apply binds_first.

      apply binds_other; eauto.
      apply IHvariant_env.
      inversion Hb.
      rewrite eq_atom_false in H3; eauto.

    destruct (x == f); subst.
      inversion Hb.
      rewrite eq_atom_true in H3.
      inversion H3; subst.
      contradict Hr; tauto.      

      apply IHvariant_env.
      inversion Hb.
      rewrite eq_atom_false in H3; eauto.
Qed.

Lemma method_present : (*$\label{method_present}$*)
  forall (a:ann) (C C0:cname) (m:mname) (Dys:a_env) (t:a_exp)
    (Dys':env) (t':exp),
    a ->
    AT C ->
    a_method C m a (C0,Dys,t) ->
    variant_env Dys Dys' ->
    variant_term t t' ->
    method C m (C0,Dys',t').
Proof.
  intros a C C0 m Dys t Dys' t' Ha HrC Hm.
  change (variant_env (snd(fst(C0,Dys,t))) Dys' ->
    variant_term (snd(C0,Dys,t)) t' ->
    method C m ((fst(fst(C0,Dys,t))),Dys',t')).
  induction Hm; intros Hve Hvt; clear C0; clear Dys; clear t;
    assert (exists fs':flds, exists ms':mths, variant_env fs fs' /\
      variant_methods C ms ms' /\ binds C (D,fs',ms') CT)
      as [fs' [ms' [_ [Hvm Hb]]]]
      by (apply class_present with (gs:=gs) (hs:=hs) (ts:=ts); eauto).
    destruct mdecl as [[C0 Dys] t]; simpl in *.
    apply method_this with (D:=D) (fs:=fs') (ms:=ms'); auto.
    clear H Hb.
    induction Hvm.
      discriminate H0.

      destruct (m0 == m); subst.
        inversion H0.
        rewrite eq_atom_true in H5.
        inversion H5; subst.
        assert (Cxs' = Dys')
          by (apply variant_env_unique with (fs:=Dys); auto).
        assert (t'0 = t')
          by (apply variant_term_unique with (t:=t); auto).
        subst; apply binds_first.

        inversion H0.
        rewrite eq_atom_false in H5; auto.
        apply binds_other; auto.

      destruct (m0 == m); subst.
        contradict H1; tauto.

        inversion H0.
        rewrite eq_atom_false in H3; auto.
      
    destruct mdecl as [[C0 Dys] t]; simpl in *.
    assert (HrCD:AT C -> AT D) by (apply class_superclass with
      (C:=C) (gs:=gs) (hs:=hs) (ts:=ts) (fs:=fs) (ms:=ms); eauto).
    apply method_super with (D:=D) (fs:=fs') (ms:=ms'); auto.
    set aCT_well_typed as Hct.
    destruct Hct as [_ Hfa].
    induction Hfa; try discriminate H.
    destruct (x == C); subst.
      inversion H.
      rewrite eq_atom_true in H4.
      inversion H4; subst; clear H4.
      destruct H2 as [_ [_ [Hokms _]]].
      revert H0 H1 Hokms Ha Hvm; clear.
      induction 5.
        discriminate H0.

        destruct (m0 == m); subst.
          contradict H3; auto.

          inversion H0.
          inversion Hokms; subst; clear Hokms.
          rewrite eq_atom_false in H5; auto.
          unfold no_binds in *; simpl.
          rewrite eq_atom_false; auto.
    
        destruct (m0 == m); subst.
          inversion Hokms; subst; clear Hokms.
          apply no_binds_ms with (C:=C) (ms:=ms); auto.

          inversion H0.
          inversion Hokms; subst.
          rewrite eq_atom_false in H3; auto.

      inversion H.
      rewrite eq_atom_false in H4; auto.

    destruct mdecl as [[C0 Dys] t]; simpl in *.
    assert (HrCD:AT C -> AT D)
      by (apply class_superclass with (C:=C) (gs:=gs) (hs:=hs)
         (ts:=ts) (fs:=fs) (ms:=ms); eauto).
    apply method_super with (D:=D) (fs:=fs') (ms:=ms'); auto.
    apply no_binds_ms with (C:=C) (ms:=ms); auto.
Qed.

Lemma method_present' : (*$\label{method_present'}$*)
  forall (a:ann) (C C0:cname) (m:mname) (Dys:a_env) (t:a_exp),
    a ->
    AT C ->
    a_method C m a (C0,Dys,t) ->
    exists Dys':env, exists t':exp,
      variant_env Dys Dys' /\
      variant_term t t' /\
      method C m (C0,Dys',t').
Proof.
  intros a C C0 m Dys t Ha HrC Hm.
  assert (HDys:exists Dys':env, variant_env Dys Dys')
    by (apply variant_env_total; auto).
  assert (Ht:exists t':exp, variant_term t t')
    by (apply variant_term_total; auto).
  destruct HDys as [Dys' Hve].
  destruct Ht as [t' Hvt].
  exists Dys'; exists t'.
  repeat split; auto.
  apply method_present with (a:=a) (Dys:=Dys) (t:=t); auto.
Qed.

Axiom equalsTrue : prop_extensionality.

Lemma method_present_reverse : (*$\label{method_present_reverse}$*)
  forall (a:ann) (C C0:cname) (m:mname) (Dys':env) (t':exp),
    method C m (C0,Dys',t') ->
    a ->
    exists Dys:a_env, exists t:a_exp,
      variant_env Dys Dys' /\
      variant_term t t' /\
      a_method C m a (C0,Dys,t).
Proof.
  intros a C C0 m Dys' t' Hm Ha.
  change ( exists Dys:a_env, exists t:a_exp,
      variant_env Dys (snd (fst (C0,Dys',t'))) /\
      variant_term t (snd (C0,Dys',t')) /\
      a_method C m a ((fst (fst (C0,Dys',t'))),Dys,t)).
  induction Hm; rename fs into fs'; rename ms into ms';
    clear C0; clear Dys'; clear t';
    destruct mdecl as [[C0 Dys'] t']; simpl in *;
    assert (exists gs:a_flds, exists hs:a_flds, exists ts:a_flds,
      exists fs:a_flds, exists ms:a_mths, variant_env fs fs' /\
      variant_methods C ms ms' /\ binds C (AT C,D,gs,hs,ts,fs,ms) aCT)
      as [gs [hs [ts [fs [ms [Hve [Hvm Hb]]]]]]]
      by (apply class_present_reverse; eauto).
    cut (exists Dys:a_env, exists t:a_exp,
      variant_env Dys Dys' /\ variant_term t t' /\
      (binds m (ATm C m,(C0,Dys,t)) ms /\ (a -> ATm C m))).
      intro Hee.
      destruct Hee as [Dys [t [Hve2 [Hvt2 [Hbm Hr]]]]].
      exists Dys; exists t.
      repeat split; auto.
      apply a_method_this with (D:=D) (fs:=fs) (gs:=gs) (hs:=hs)
        (ts:=ts) (ms:=ms); auto.

      revert H0 Ha Hvm; clear.
      induction 3; try discriminate H0.
        inversion H0.
        destruct (m0 == m); subst.
          rewrite eq_atom_true in H4; inversion H4; subst; clear H4.
          exists Cxs; exists t.
          repeat split; auto.
          apply binds_first.

          rewrite eq_atom_false in H4; auto.
          destruct IHHvm as [Dys [t1 [Hve [Hvt [Hb Hr]]]]]; auto.
          exists Dys; exists t1.
          repeat split; auto.
          apply binds_other; auto.

        destruct IHHvm as [Dys [t1 [Hve [Hvt [Hb Hr]]]]]; auto.
        exists Dys; exists t1.
        repeat split; auto.
        destruct (m0 == m); subst.
          contradict H; auto.

          apply binds_other; auto.

    cut (exists Dys:a_env, exists t:a_exp,
      variant_env Dys Dys' /\ variant_term t t' /\
      ((exists mdecl':typ*a_env*a_exp, binds m (ATm C m,mdecl') ms /\
        ~(a -> ATm C m) /\ a_method D m (a /\ ~ATm C m) (C0,Dys,t)) \/
      no_binds m ms /\ a_method D m a (C0,Dys,t))).
      intro Hee.
      destruct Hee as [Dys [t [Hve2 [Hvt2 [Hnotthis | Hsuper]]]]];
        exists Dys; exists t; repeat split; auto.
        destruct Hnotthis as [mdecl' [Hbm [Hr Hsuper]]].
        apply a_method_notthis with (D:=D) (fs:=fs) (gs:=gs) (hs:=hs)
          (ts:=ts) (ms:=ms) (mdecl':=mdecl'); auto.

        destruct Hsuper as [Hnbm Hsuper].
        apply a_method_super with (D:=D) (fs:=fs) (gs:=gs) (hs:=hs)
          (ts:=ts) (ms:=ms); auto.

      revert H0 Ha IHHm Hvm; clear.
      induction 4.
        destruct IHHm as [Dys [t1 [Hve [Hvt Hsuper]]]]; exists Dys;
          exists t1; repeat split; auto.
        right; split; try apply nobinds_nil; auto.

        inversion H0.
        destruct (m0 == m); subst.
          rewrite eq_atom_true in H4; inversion H4.

          rewrite eq_atom_false in H4; auto.
          destruct IHHvm as [Dys [t1 [Hve [Hvt [Hnotthis | Hsuper]]]]
            ]; auto; exists Dys; exists t1; repeat split; auto.
            destruct Hnotthis as [mdecl' [Hb [Hr Hsuper]]].
            left; exists mdecl'; split; auto.
            apply binds_other; auto.

            destruct Hsuper as [Hnbm Hsuper].
            right; split; auto.
            apply nobinds_cons; auto.

        destruct (m0 == m); subst.
          destruct IHHm as [Dys [t1 [Hve [Hvt Hsuper]]]]; exists Dys;
            exists t1; repeat split; auto.
          left; exists (C1,Cxs,t); repeat split; auto.
            apply binds_first.

            revert Ha H Hsuper; clear; intros.
            assert (He:a=True) by (apply equalsTrue; tauto).
            assert (He':(a /\ ~ATm C m)=True) by (apply equalsTrue; tauto).
            rewrite He in *; rewrite He'; eauto.

          destruct IHHvm as [Dys [t1 [Hve [Hvt [Hnotthis | Hsuper]]]]
            ]; auto; exists Dys; exists t1; repeat split; auto.
            destruct Hnotthis as [mdecl' [Hb [Hr Hsuper]]].
            left; exists mdecl'; split; auto.
            apply binds_other; auto.

            destruct Hsuper as [Hnbm Hsuper].
            right; split; auto.
            apply nobinds_cons; auto.
Qed.

Lemma type_present : (*$\label{type_present}$*)
  forall (a:ann) (E:a_env) (t:a_exp) (C:typ),
    a_typing a E t C -> (a -> AT C).
Proof.
  induction 1; auto.
Qed.

Lemma subtype_present : (*$\label{subtype_present}$*)
  forall (C D:typ),
    AT C ->
    a_sub C D ->
    sub C D.
Proof.
  induction 2; eauto.
    cut (AT t2); eauto.
    revert H H0_; clear.
    induction 2; eauto.
    destruct H0 as [gs [hs [ts [fs [ms Hb]]]]].
    apply (class_superclass C D gs hs ts fs ms); auto.

    destruct H0 as [gs [hs [ts [fs [ms Hb]]]]].
    destruct (class_present C D gs hs ts fs ms)
      as [fs' [ms' [_ [_ Hb']]]]; eauto.
Qed.
