Require Import Metatheory.
Require Import FJ_Definitions.
Require Import CFJ_Definitions.
Require Import CFJ_Properties.
Require Import Coq.Logic.Classical_Prop.

(********************************************************************)
(* AUXILIARY PROOFS                                                 *)
(********************************************************************)

Fact equal : forall a:Prop, a <-> a.
Proof.
  intro; split; auto.
Qed.
Hint Resolve nobinds_nil nobinds_cons equal.

(********************************************************************)
(* CFJ PRODUCT LINE                                                 *)
(********************************************************************)

(* The features. The base feature representing the code not annotated
   with other features do not needs to be represented here. *)
Variables Second : ann.
Variables SecondCall : ann.

(* The feature model. *)
Hypothesis fm_fix : FM = (Second <-> SecondCall).

(* Fresh class names. *)
Variables A B Pair Library : cname.
Hypothesis A_fresh: Object <> A.
Hypothesis B_fresh: Object <> B /\ A <> B.
Hypothesis Pair_fresh : Object <> Pair /\ A <> Pair /\ B <> Pair.
Hypothesis Library_fresh : Object <> Library /\ A <> Library /\
  B <> Library /\ Pair <> Library.

(* Auxiliaries. *)
Definition pairx (a:ann * a_exp) (b:ann * a_exp) :=
  ae_new Pair (a :: b :: nil).
Definition pair'x (a:exp) (b:exp) :=
  e_new Pair (a :: b :: nil).

(* Fields of Pair. *)
Variables fst snd : fname.
Hypothesis snd_fresh : fst <> snd.

Definition pair_flds : a_flds :=
  (fst,(FM,Object)) ::
  (snd,(FM /\ Second,Object)) ::
  nil.

(* Methods of Library. *)
Variable second : var.
Hypothesis second_fresh : this <> second.
Variable pairAX : mname.
Definition pairAX_env : a_env := (second,(FM,Object)) :: nil.
Definition pairAX_body : a_exp :=
  ae_new Pair ((FM,ae_new A nil) ::
               (FM /\ SecondCall,ae_var second) ::
               nil).

Variable first : var.
Hypothesis first_fresh : this <> first.
Variable pairXA : mname.
Hypothesis pairXA_fresh : pairAX <> pairXA.
Definition pairXA_env : a_env := (first,(FM,Object)) :: nil.
Definition pairXA_body : a_exp :=
  ae_new Pair ((FM,ae_var first) :: 
               (FM /\ SecondCall,ae_new A nil) ::
               nil).

Definition library_mths : a_mths :=
  (pairAX,(FM,(Pair,pairAX_env,pairAX_body))) ::
  (pairXA,(FM,(Pair,pairXA_env,pairXA_body))) ::
  nil.

(* The CFJ class table. *)
Hypothesis act_fix : aCT =
    (A,(FM,Object,nil,nil,nil,nil,nil)) ::
    (B,(FM,Object,nil,nil,nil,nil,nil)) ::
    (Pair,(FM,Object,pair_flds,nil,pair_flds,pair_flds,nil)) :: 
    (Library,(FM,Object,nil,nil,nil,nil,library_mths)) :: 
    nil.

(* The CFJ start term. *)
Definition aterm :=
  ae_field (
    ae_meth (
      ae_new Library nil
    ) pairAX ((FM,ae_new B nil)::nil)
  ) fst.

Hint Unfold pair_flds pairAX_env pairAX_body pairXA_env pairXA_body
  library_mths act_fix aterm.
Hint Resolve fm_fix.

(********************************************************************)
(* FACTS ON THE CFJ PRODUCT LINE                                    *)
(********************************************************************)

(* Facts on fresh variable names. *)
Lemma A_fresh_not: ~(Object = A).
Proof.
  exact A_fresh.
Qed.
Lemma B_fresh_not: ~(Object = B \/ A = B).
Proof.
  destruct B_fresh; tauto.
Qed.
Lemma Pair_fresh_not: ~(Object = Pair \/ A = Pair \/ B = Pair).
Proof.
  destruct Pair_fresh as [HO [HA HB]]; tauto.
Qed.
Lemma Library_fresh_not: ~(Object = Library \/ A = Library \/
  B = Library \/ Pair = Library).
Proof.
  destruct Library_fresh as [HO [HA [HB HPair]]]; tauto.
Qed.
Lemma snd_fresh_not: ~(fst = snd).
Proof.
  exact snd_fresh.
Qed.
Lemma pairXA_fresh_not: ~(pairAX = pairXA).
Proof.
  exact pairXA_fresh.
Qed.
Lemma second_fresh_not: ~(this = second).
Proof.
  exact second_fresh.
Qed.
Lemma first_fresh_not: ~(this = first).
Proof.
  exact first_fresh.
Qed.
Hint Resolve A_fresh_not B_fresh_not Pair_fresh_not
  Library_fresh_not snd_fresh_not pairXA_fresh_not second_fresh_not
  first_fresh_not.

(* Annotation table values. *)
Fact object_present : AT Object = True.
Proof.
  unfold AT; rewrite act_fix; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_true; auto.
Qed.
Fact object_present' : AT Object.
Proof.
  rewrite object_present; trivial.
Qed.
Fact A_present : AT A = FM.
Proof.
  unfold AT; rewrite act_fix; simpl.
  rewrite eq_atom_true; auto.
Qed.
Fact A_present' : FM -> AT A.
Proof.
  rewrite A_present; trivial.
Qed.
Fact B_present : AT B = FM.
Proof.
  unfold AT; rewrite act_fix; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_true; auto.
Qed.
Fact B_present' : FM -> AT B.
Proof.
  rewrite B_present; trivial.
Qed.
Fact Pair_present : AT Pair = FM.
Proof.
  unfold AT; rewrite act_fix; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_true; auto.
Qed.
Fact Pair_present' : FM -> AT Pair.
Proof.
  rewrite Pair_present; trivial.
Qed.
Fact Library_present : AT Library = FM.
Proof.
  unfold AT; rewrite act_fix; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto 6.
  rewrite eq_atom_false; auto 6.
  rewrite eq_atom_true; auto.
Qed.
Fact Library_present' : FM -> AT Library.
Proof.
  rewrite Library_present; trivial.
Qed.
Fact pairAX_present : ATm Library pairAX = FM.
Proof.
  unfold ATm; rewrite act_fix; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto 6.
  rewrite eq_atom_false; auto 6.
  rewrite eq_atom_true; auto.
  unfold library_mths; simpl.
  rewrite eq_atom_true; auto.
Qed.
Fact pairAX_present' : FM -> ATm Library pairAX.
Proof.
  rewrite pairAX_present; trivial.
Qed.
Fact pairXA_present : ATm Library pairXA = FM.
Proof.
  unfold ATm; rewrite act_fix; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto 6.
  rewrite eq_atom_false; auto 6.
  rewrite eq_atom_true; auto.
  unfold library_mths; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_true; auto.
Qed.
Fact pairXA_present' : FM -> ATm Library pairXA.
Proof.
  rewrite pairXA_present; trivial.
Qed.
Fact fst_present : ATf Pair fst = FM.
Proof.
  unfold ATf; rewrite act_fix; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_true; auto.
  unfold pair_flds; simpl.
  rewrite eq_atom_true; auto.
Qed.
Fact pfst_present : ATp Pair fst = FM.
Proof.
  unfold ATp; rewrite act_fix; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_true; auto.
  unfold pair_flds; simpl.
  rewrite eq_atom_true; auto.
Qed.
Fact afst_present : ATa Pair fst = FM.
Proof.
  unfold ATa; rewrite act_fix; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_true; auto.
  unfold pair_flds; simpl.
  rewrite eq_atom_true; auto.
Qed.
Fact fst_present' : FM -> ATf Pair fst.
Proof.
  rewrite fst_present; trivial.
Qed.
Fact snd_present : ATf Pair snd = (FM /\ Second).
Proof.
  unfold ATf; rewrite act_fix; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_true; auto.
  unfold pair_flds; rewrite fm_fix; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_true; auto.
Qed.
Fact psnd_present : ATp Pair snd = (FM /\ Second).
Proof.
  unfold ATp; rewrite act_fix; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_true; auto.
  unfold pair_flds; rewrite fm_fix; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_true; auto.
Qed.
Fact asnd_present : ATa Pair snd = (FM /\ Second).
Proof.
  unfold ATa; rewrite act_fix; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_true; auto.
  unfold pair_flds; rewrite fm_fix; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_true; auto.
Qed.
Hint Resolve object_present Pair_present
  pairAX_present pairXA_present pairAX_present' pairXA_present'
  fst_present afst_present pfst_present fst_present'
  snd_present psnd_present asnd_present object_present' Pair_present'
  A_present A_present' B_present B_present'
  Library_present Library_present'.

(* Bindings to the class table. *)
Fact object_notin_ct : forall x, ~(binds Object x aCT).
Proof.
  intro; rewrite act_fix; unfold binds; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto.
  intuition; discriminate H.
Qed.
Fact binds_A : binds A (AT A, Object, nil, nil, nil, nil, nil) aCT.
Proof.
  rewrite act_fix; unfold binds; simpl.
  rewrite eq_atom_true.
  rewrite A_present; auto.
Qed.
Fact binds_B : binds B (AT B, Object, nil, nil, nil, nil, nil) aCT.
Proof.
  rewrite act_fix; unfold binds; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_true.
  rewrite B_present; auto.
Qed.
Fact binds_Library : binds Library (AT Library, Object, nil, nil, nil,
  nil, library_mths) aCT.
Proof.
  rewrite act_fix; unfold binds; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto 6.
  rewrite eq_atom_false; auto 6.
  rewrite eq_atom_true.
  rewrite Library_present; auto.
Qed.
Fact binds_fst : binds fst (ATf Pair fst, Object) pair_flds.
Proof.
  unfold pair_flds; unfold binds; simpl.
  rewrite eq_atom_true.
  rewrite fst_present; auto.
Qed.
Fact binds_pairAX : binds pairAX (ATm Library pairAX, (Pair,
  pairAX_env, pairAX_body)) library_mths.
Proof.
  unfold library_mths; unfold binds; simpl.
  rewrite eq_atom_true.
  rewrite pairAX_present; auto.
Qed.
Fact fields_Pair : a_fields Pair pair_flds.
Proof.
  replace pair_flds with (nil++pair_flds); auto.
  apply a_fields_other with (D:=Object) (gs:=pair_flds)
    (hs:=nil) (ts:=pair_flds) (ms:=nil); auto.
  rewrite act_fix.
  unfold binds; simpl.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_false; auto.
  rewrite eq_atom_true.
  replace (AT Pair) with FM; auto.
Qed.
Hint Resolve object_notin_ct binds_A binds_B binds_Library binds_fst
  binds_pairAX fields_Pair.

(********************************************************************)
(* CFJ PRODUCT LINE IS WELL-TYPED                                   *)
(********************************************************************)

Theorem aCT_well_typed : a_ok_ctable aCT.
Proof.
  rewrite act_fix.
  unfold a_ok_ctable; split.
    auto 10.

    repeat apply fa_cons; auto; unfold a_ok_class';
      unfold a_ok_class;
      (repeat split; [
        exists nil ; repeat split; simpl; auto | 
        auto | 
        auto | 
        auto]).
      unfold library_mths.
      rewrite <- pairAX_present at 1.
      rewrite <- pairXA_present at 1.
      repeat apply a_occ_cons_mths; try apply a_occ_nil;
        rewrite Library_present; try rewrite pairAX_present;
        try rewrite pairXA_present; auto.

      unfold library_mths; auto.

      unfold library_mths.
      rewrite <- pairAX_present at 1.
      rewrite <- pairXA_present at 1.
      apply fa_cons.
        apply fa_cons; auto.
        unfold a_ok_meth'; unfold a_ok_meth.
        unfold pairXA_env; unfold pairXA_body.
        repeat split; auto.
          inversion H; subst; contradict H0; auto.

          inversion H; subst; contradict H0; auto.

          apply a_wt_sub with (t:=Pair); auto.
          apply a_t_new with (Dfs:=pair_flds); auto.
          unfold pair_flds; simpl.
          apply a_wts_cons; auto.
            apply a_wts_cons; try apply a_wts_nil;
              auto using nobinds_nil, nobinds_cons.
                apply a_wt_sub with (t:=A); eauto 8.
                apply a_t_new with (Dfs:=nil); simpl; auto.
                  apply a_fields_other with (fs':=nil) (fs:=nil)
                    (D:=Object) (gs:=nil) (hs:=nil) (ts:=nil)
                    (ms:=nil); auto.

                  intro H; destruct H; auto.

                rewrite pairXA_present.
                rewrite fm_fix; tauto.

              destruct 1; auto.

            apply a_wt_sub with (t:=Object); auto.
            apply a_t_var with (a':=FM); auto.
            unfold binds; simpl.
            rewrite eq_atom_false; auto.
            rewrite eq_atom_true; auto.

        rewrite Pair_present; rewrite pairXA_present; auto.

        unfold a_ok_meth'; unfold a_ok_meth.
        unfold pairAX_env; unfold pairAX_body.
        repeat split; auto.
          inversion H; subst; contradict H0; auto.

          inversion H; subst; contradict H0; auto.

          apply a_wt_sub with (t:=Pair); auto.
          apply a_t_new with (Dfs:=pair_flds); auto.
            unfold pair_flds; simpl.
            apply a_wts_cons; auto.
              apply a_wts_cons; try apply a_wts_nil; auto.
                apply a_wt_sub with (t:=Object); auto.
                apply a_t_var with (a':=FM); auto.
                  unfold binds; simpl.
                  rewrite eq_atom_false; auto.
                  rewrite eq_atom_true; auto.

                  destruct 1; auto.

                rewrite pairAX_present; rewrite fm_fix; tauto.

                rewrite pairAX_present; rewrite fm_fix; tauto.

                apply a_wt_sub with (t:=A); eauto 8.

            rewrite pairAX_present; rewrite Pair_present; auto.

      unfold pair_flds.
      rewrite <- fst_present at 1.
      rewrite <- snd_present at 1.
      apply a_occ_cons_flds1; auto.
        apply a_occ_cons_flds1; auto.
          rewrite snd_present; intro; destruct H.
          rewrite Pair_present; rewrite object_present; split; auto.

          rewrite asnd_present; rewrite psnd_present.
          rewrite snd_present; tauto.

        rewrite fst_present.
        rewrite Pair_present; rewrite object_present; split; auto.

        rewrite afst_present; rewrite pfst_present.
        rewrite fst_present; tauto.

      unfold pair_flds; auto.
Qed.

Theorem product_line_well_typed : FM -> CFJ_product_line aterm.
Proof.
  unfold CFJ_product_line.
  split.
    apply aCT_well_typed.

    exists Object; eauto.
    unfold aterm.
    apply a_t_field with (C0:=Pair); eauto.
    apply a_t_meth with (Dys:=pairAX_env) (t:=pairAX_body)
      (C0:=Library); eauto.
    unfold pairAX_env; simpl; eauto 15.
Qed. 

(********************************************************************)
(* CONFIGURATION AND FJ PROGRAM                                     *)
(********************************************************************)

(* The configuration. *)
Hypothesis Second_selected : Second = False.
Hypothesis SecondCall_selected : SecondCall = False.

(* The FJ class table. *)
Definition pair_flds' : flds :=
  (fst,Object) ::
  (*(snd,Object) ::*)
  nil.

Definition pairAX_env' : env := (second,Object) :: nil.
Definition pairAX_body' : exp :=
  e_new Pair ((e_new A nil) ::
              (*(e_var second)::*)
              nil).
Definition pairXA_env' : env := (first,Object) :: nil.
Definition pairXA_body' : exp :=
  e_new Pair ((e_var first) ::
              (*(e_new A nil)::*)
              nil).

Definition library_mths' : mths :=
  (pairAX,(Pair,pairAX_env',pairAX_body')) ::
  (pairXA,(Pair,pairXA_env',pairXA_body')) ::
  nil.

Hypothesis ct_fix : CT =
    (A,(Object,nil,nil)) ::
    (B,(Object,nil,nil)) ::
    (Pair,(Object,pair_flds',nil)) :: 
    (Library,(Object,nil,library_mths')) :: 
    nil.

Definition term :=
  e_field (
    e_meth (
      e_new Library nil
    ) pairAX ((e_new B nil)::nil)
  ) fst.

Hint Unfold pair_flds' pairAX_env' pairAX_body' pairXA_env'
  pairXA_body' library_mths' Second_selected ct_fix term.

Fact Second_selected' : ~ Second.
Proof.
  rewrite Second_selected; auto.
Qed.
Fact SecondCall_selected' : ~ SecondCall.
Proof.
  rewrite SecondCall_selected; auto.
Qed.
Fact Second_SecondCall : Second <-> SecondCall.
Proof.
  rewrite Second_selected; rewrite SecondCall_selected; tauto.
Qed.

Hint Resolve Second_selected' SecondCall_selected' Second_SecondCall.

(********************************************************************)
(* VARIANT GENERATION PRODUCES FJ PROGRAM                           *)
(********************************************************************)

(* CT is the variant generated from aCT *)
Theorem aCT_correlates_with_CT : variant_ct aCT CT.
Proof.
  rewrite act_fix; rewrite ct_fix.
  rewrite <- A_present at 1.
  rewrite <- B_present at 1.
  rewrite <- Pair_present at 1.
  rewrite <- Library_present at 1.
  apply vct_cons_y; try rewrite A_present; try rewrite fm_fix; auto.
  apply vct_cons_y; try rewrite B_present; try rewrite fm_fix; auto.
  apply vct_cons_y; try rewrite Pair_present; try rewrite fm_fix;auto.
    apply vct_cons_y; try rewrite Library_present; try rewrite fm_fix;
      auto.
    unfold library_mths; unfold library_mths'.
    rewrite <- pairAX_present at 1.
    rewrite <- pairXA_present at 1.
    apply vms_cons_y; try rewrite pairAX_present; try rewrite fm_fix;
      auto.
      apply vms_cons_y; try rewrite pairXA_present;
        try rewrite fm_fix; auto.
        unfold pairXA_body; unfold pairXA_body'.
        apply vt_new.
        apply vts_cons_y; try rewrite fm_fix; auto.
        apply vts_cons_n; try rewrite Second_selected; auto;intuition.

        unfold pairXA_env; unfold pairXA_env'.
        apply ve_cons_y; try rewrite fm_fix; auto.

      unfold pairAX_body; unfold pairAX_body'.
      apply vt_new.
      apply vts_cons_y; try rewrite fm_fix; auto.
      apply vts_cons_n; try rewrite Second_selected; auto; intuition.

      unfold pairAX_env; unfold pairAX_env'.
      apply ve_cons_y; try rewrite fm_fix; auto.

    unfold pair_flds; unfold pair_flds'.
    apply ve_cons_y; try rewrite fm_fix; auto.
    apply ve_cons_n; try rewrite Second_selected; auto; intuition.
Qed.

Theorem aterm_correlates_with_term : variant_term aterm term.
Proof.
  unfold aterm; unfold term.
  apply vt_field.
  apply vt_meth; auto.
  apply vts_cons_y; try rewrite fm_fix; auto.
Qed.

Theorem FJ_program_well_typed : FJ_program term.
Proof.
  apply generation_preserves_typing with (t:=aterm).
    apply product_line_well_typed.
    rewrite fm_fix; auto.

    apply aCT_correlates_with_CT.

    apply aterm_correlates_with_term.

    rewrite fm_fix; auto.
Qed.
