Require Import Metatheory.
Require Import FJ_Definitions.
Require Import CFJ_Definitions.
Require Import Coq.Logic.Classical_Prop.

(********************************************************************)
(* AUXILIARY PROOFS                                                 *)
(********************************************************************)

Fact no_binds_env : (*$\label{no_binds_env}$*)
  forall (x:var) (E:a_env) (E':env),
    no_binds x E ->
    variant_env E E' ->
    no_binds x E'.
Proof.
  intros x E E' Hnb Hve.
  induction Hve.
    apply nobinds_nil.

    unfold no_binds in *.
    simpl in Hnb.
    destruct (x == x0).
      discriminate Hnb.

      apply nobinds_cons; try apply IHHve; try assumption.

    apply IHHve.
    unfold no_binds in *.
    simpl in Hnb.
    destruct (x == x0).
      discriminate Hnb.

      assumption.
Qed.

Fact ok_env : (*$\label{ok_env}$*)
  forall (E:a_env) (E':env),
    ok E ->
    variant_env E E' ->
    ok E'.
Proof.
  intros E E' Hok Hve.
  induction Hve.
    apply ok_nil.

    apply ok_cons.
      apply IHHve; inversion Hok; assumption.

      apply no_binds_env with (E:=E).
        inversion Hok; assumption.

        assumption.

    apply IHHve; inversion Hok; assumption.
Qed.

Fact ok_env_env : (*$\label{ok_env_env}$*)
  forall (E F:a_env) (E' F':env),
    ok (E ++ F) ->
    variant_env E E' ->
    variant_env F F' ->
    ok (E' ++ F').
Proof.
  intros E F E' F' Hok HveE; revert F'.
  induction HveE; intros.
    simpl in Hok.
    simpl.
    apply ok_env with (E:=F); assumption.

    rewrite <- app_comm_cons in Hok.
    rewrite <- app_comm_cons.
    inversion_clear Hok; subst.
    apply ok_cons; eauto.
    assert (Hsplit:no_binds x E /\ no_binds x F).
      revert H2; clear.
      unfold no_binds. intros.
      induction E as [|(y,v)]; simpl in *; split; eauto;
        (destruct (x == y); [discriminate H2 | apply IHE; eauto]).

      destruct Hsplit.
      apply nobinds_app.
        apply no_binds_env with (E:=E); try assumption.

        apply no_binds_env with (E:=F); try assumption.

    rewrite <- app_comm_cons in Hok.
    inversion Hok; apply IHHveE; assumption.
Qed.

Fact no_binds_ms : (*$\label{no_binds_ms}$*)
  forall (C:cname) (m:mname) (ms:a_mths) (ms':mths),
    no_binds m ms ->
    variant_methods C ms ms' ->
    no_binds m ms'.
Proof.
  unfold no_binds.
  intros.
  induction H0; trivial.
    simpl.
    simpl in H.
    destruct (m == m0).
      discriminate H.

      apply IHvariant_methods; assumption.

    apply IHvariant_methods.
    inversion H; clear H; subst.
    destruct (m == m0).
      discriminate H3.

      trivial.
Qed.

Fact no_binds_ct : (*$\label{no_binds_ct}$*)
  forall (C:cname) (ct:a_ctable) (ct':ctable),
    no_binds C ct ->
    variant_ct ct ct' ->
    no_binds C ct'.
Proof.
  intros C ct ct' Hnb Hvct.
  induction Hvct.
    apply nobinds_nil.

    unfold no_binds in *.
    simpl in Hnb.
    destruct (C == C0).
      discriminate Hnb.

      apply nobinds_cons; try apply IHHvct; assumption.

    apply IHHvct.
    unfold no_binds in *.
    simpl in Hnb.
    destruct (C == C0).
      discriminate Hnb.

      assumption.
Qed.

Fact vtu_mut : (*$\label{vtu_mut}$*)
  ( forall (t:a_exp) (t':exp),
    variant_term t t' ->
    forall (t'':exp), variant_term t t'' -> t' = t'' ) /\
  ( forall (ts:list (ann*a_exp)) (ts':list exp),
    variant_terms ts ts' ->
    forall (ts'':list exp), variant_terms ts ts'' -> ts' = ts'' ).
Proof.
  apply variant_term_mutind.
    inversion 1; intuition (f_equal; eauto).
    inversion 3; intuition (f_equal; eauto).
    inversion 5; intuition (f_equal; eauto).
    inversion 3; intuition (f_equal; eauto).
(* EDIT4 BEGIN *)
    inversion 3; subst.
      apply H0 in H5; inversion H5; intuition.

      apply H0 in H4; discriminate H4.
    inversion 5; subst.
      apply H0 in H7; discriminate H7.

      apply H2 in H8; inversion H8; intuition.
(* EDIT4 END *)
    inversion 1; intuition (f_equal; eauto).
    inversion 6; intuition (f_equal; eauto).
    inversion 4; intuition (f_equal; eauto).
Qed.

Fact variant_term_unique : (*$\label{variant_term_unique}$*)
  forall (t:a_exp) (t' t'':exp),
    variant_term t t' ->
    variant_term t t'' ->
    t' = t''.
Proof.
  destruct vtu_mut; eauto.
Qed.

Fact variant_env_unique : (*$\label{variant_env_unique}$*)
  forall (fs:a_flds) (fs' fs'':flds),
    variant_env fs fs' ->
    variant_env fs fs'' ->
    fs' = fs''.
Proof.
  intros fs fs' fs'' H; revert fs''.
  induction H; inversion 1; intuition (f_equal; eauto).
Qed.

Definition vtt_trick (*$\label{vtt_trick}$*)
  (REC : forall (t:a_exp), exists t':exp, variant_term t t') :
  forall (ts:list (ann*a_exp)),
    exists ts':list exp, variant_terms ts ts'.
Proof.
  intro H.
  induction ts.
    exists nil; auto.

    destruct a as [a t].
    destruct H with (t:=t) as [t' Hvt].
    destruct IHts as [ts' Hts].
    assert (Ha:a \/ ~a) by apply classic.
    destruct Ha.
      exists (t'::ts'); auto.

      exists ts'; auto.
Defined.

Fact variant_term_total : (*$\label{variant_term_total}$*)
  forall (t:a_exp), exists t':exp, variant_term t t'.
Proof.
  refine (fix variant_term_total (t : a_exp) := _); destruct t;
    try destruct (variant_term_total t) as [t' Hvt];
    try destruct (vtt_trick variant_term_total l) as [l' Hvts].
    exists (e_var v); auto.

    exists (e_field t' f); auto.

    exists (e_meth t' m l'); auto.

    exists (e_new c l'); auto.
(* EDIT4 BEGIN *)

    destruct l'.
      exists t'; eauto.

      exists e; eauto.
(* EDIT4 END *)
Qed.

Fact variant_env_total : (*$\label{variant_env_total}$*)
  forall (fs:a_flds),
    exists fs':flds, variant_env fs fs'.
Proof.
  intros fs.
  induction fs.
    exists nil; auto.

    destruct a as [f [a C]].
    destruct IHfs as [fs' Hve].
    assert (Ha:a \/ ~a) by apply classic.
    destruct Ha.
      exists ((f, C)::fs'); auto.

      exists fs'; auto.
Qed.
